# Future Library
from __future__ import annotations

# Standard Library
# Python Modules
import datetime
import importlib
import logging

from dataclasses import MISSING, Field, fields, is_dataclass
from types import NoneType
from typing import (
    Any,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Self,
    Type,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

# 3rd Party Library
# 3rd Party Modules
from dateutil.parser import parse as parse_date
from icontract import ViolationError, require
from pydantic.dataclasses import dataclass

# 1st Party Library
# Project Modules
from rwskit import types_
from rwskit.types_ import type_from_string

log = logging.getLogger(__name__)

D = TypeVar("D")
"""A type intended to represent a dataclass."""


# I can't get the typing right. Given the type hints below, the code executes
# properly at runtime, but pylance cannot determine constructor parameters
# for any class that uses this decorator. Oddly, directly using the
# pydantic dataclass decorator works just fine.
# def immutable_dataclass(cls: Type[D]) -> Type[D]:
#     """A decorator to convert a class to a frozen keyword only pydantic dataclass."""
#
#     return cast(
#         Type[D],
#         dataclass(
#             cls,
#             frozen=True,
#             kw_only=True,
#             config={"arbitrary_types_allowed": True},
#         ),
#     )


def is_frozen(cls: Type) -> bool:
    """Checks if a class is a frozen dataclass.

    This will return ``True`` if the input ``cls`` is both a dataclass (either
    ``dataclasses.dataclass`` or ``pydantic.dataclasses.dataclass``) and
    it is frozen.
    """
    if not is_dataclass(cls):
        return False

    return getattr(cls, "__dataclass_params__").frozen


def has_default_value(field: Field) -> bool:
    """
    Checks if a ``field`` has a default value.

    Parameters
    ----------
    field : Field
        The field to check.

    Returns
    -------
    bool
        ``True`` if the field has a default value either directly or via
        a default factory.
    """
    return not (field.default is MISSING and field.default_factory is MISSING)


def is_optional(field: Field) -> bool:
    """Checks if a ``field`` is optional.

    Parameters
    ----------
    field : Field
        The field to check.

    Returns
    -------
    bool
        ``True`` if the field is optional.
    """
    return types_.is_optional(field.type)


def is_required(field: Field) -> bool:
    """Checks if a ``field`` is required.

    A field is required if it is not optional and does not have a default value.

    Parameters
    ----------
    field : Field
        The field to check,

    Returns
    -------
    bool
        ``True`` if the field is required.
    """
    return not (is_optional(field) or has_default_value(field))


@require(lambda cls: is_dataclass(cls), "The class 'cls' must be a dataclass.")
def resolve_type_hints(
    cls: Type[Any], type_map: Optional[dict[str, Type[Any]]] = None, max_tries: int = 50
) -> dict[str, Any]:
    """Try to resolve the type hints for a dataclass.

    Sometimes func:`typing.get_type_hints` fails with a ``NameError`` when
    the type hint has not already been imported. This method will try to
    catch those errors and dynamically import the unresolved class.

    Parameters
    ----------
    cls : Type[Any]
        The dataclass to resolve the type hints for.
    type_map : dict[str, Type[Any]], optional
        Additional mappings from string names to type classes, which can be
        used in addition to ``globals()`` for resolving type hint classes.
    max_tries : int
        The maximum number of times the method will try to resolve missing
        type hints.

    Returns
    -------
    dict[str, Any]
        The dictionary containing the resolved type hints.

    Raises
    ------
    icontract.ViolationError
        If the input ``cls`` is not a dataclass.
    """
    # We need to ensure that any type hint used in the dataclass (including
    # ones defined in its parents) are imported and available when
    # 'get_type_hints' is called. To do that we'll get the list of classes
    # in the current classes type hierarchy and dynamically import all the
    # objects in its module. This is a very heavy-handed way to ensure the
    # necessary objects are available, but it's not obvious there is a
    # more fine-grained way to guarantee the necessary objects are available.
    class_hierarchy = cls.mro()
    dynamic_vars = [
        vars(importlib.import_module(c.__module__)) for c in class_hierarchy
    ]
    dynamic_vars = {k: v for d in dynamic_vars for k, v in d.items()}
    globalns = globals() | (type_map or {}) | dynamic_vars

    for _ in range(max_tries):
        try:
            hints = get_type_hints(cls, globalns=globalns, localns=locals())
        except NameError as e:
            # Try to resolve (and import) unknown classes.
            # The error message will look something like:
            # NameError: name 'ClassName' is not defined.
            # Splitting on `'` should give us 3 tokens with the class name in
            # the middle (with index 1).
            missing_class_name = str(e).split("'")[1]
            try:
                globalns[missing_class_name] = type_from_string(missing_class_name)
            except (ViolationError, ModuleNotFoundError):
                pass  # Not all type hints are classes.
        else:
            return hints

    raise NameError(f"Could not resolve all type hints for {cls}.")


# region Construct Dataclass
@require(
    lambda data, dataclass_type: set(data.keys()).issubset({
        f.name for f in fields(dataclass_type)
    }),
    "The 'data' keys must be a subset of the dataclass fields.",
)
@require(
    lambda data, dataclass_type: all(
        not is_required(f) or data.get(f.name) is not None
        for f in fields(dataclass_type)
    ),
    "The 'data' must contain non-null entries for all required fields.",
)
@require(
    lambda dataclass_type: is_dataclass(dataclass_type),
    "'dataclass_type' must be a dataclass.",
)
@require(
    lambda data: isinstance(data, Mapping), "'data' must be a mapping (e.g., dict)."
)
def construct_dataclass(
    data: Mapping[str, Any],
    dataclass_type: Type[D],
    type_map: Optional[dict[str, Any]] = None,
) -> D:
    """Construct a dataclass from a mapping.

    Try to construct a possibly nested dataclass of ``dataclass_type``
    from a mapping provided by ``data``. If the dataclass cannot be
    constructed from the data a ``TypeError`` will be raised.

    Parameters
    ----------
    data : Mapping[str, Any]
        The dictionary (or other mapping) to try to convert.
    dataclass_type : Type[D]
        The dataclass type we will try to construct.
    type_map : dict[str, Any], optional
        By default this method will use ``globals()`` to try to resolve the
        type hints for the dataclass fields. You can specify any additional
        mappings with ``type_map`` which will be appended to the default.

    Returns
    -------
    D
        The dataclass object we have constructed.

    Raises
    ------
    icontract.ViolationError
        If the ``dataclass_type`` is not a dataclass.
    TypeError
        If the dictionary cannot be converted to the dataclass type.

    """
    type_map = type_map or {}

    type_hints = resolve_type_hints(dataclass_type, type_map)

    # Create the keyword arguments we'll use to construct the dataclass.
    # We will only create arguments that have an entry in the mapping
    # and let the dataclass try to create defaults where necessary
    # when entries are missing in the mapping. Note, this might fail,
    # but that just means the data is not compatible with this dataclass.
    kwargs = {
        f.name: _handle_field(f, type_hints[f.name], data[f.name], type_map)
        for f in fields(cast(Any, dataclass_type))
        if f.name in data
    }

    try:
        return dataclass_type(**kwargs)
    except TypeError as e:
        raise ValueError(
            f"Error creating an instance of {dataclass_type.__name__}: {e}"
        )


def _handle_field(field, field_type: Any, value: Any, type_map: dict[str, Any]) -> Any:
    """Try to construct a dataclass field from a given value.

    Parameters
    ----------
    field : dataclasses.Field
        The dataclass field we are trying to construct.
    field_type : Any
        The type of the field. We use this value instead of ``Field.type``
        because any forward references should have already been resolved.
    value : Any
        The value to construct the field from.

    Returns
    -------
    Any
        The appropriate value for the field.

    Raises
    ------
    ValueError
        If we can't construct the field from the given value.

    """
    # Handle Union types
    if types_.is_union(field_type):
        return _handle_union_type(field, field_type, value, type_map)

    # Anything below here should be either a raw type, generic, or nested
    # dataclass.
    field_origin = get_origin(field_type)

    # Handle basic types or dataclasses.
    # This needs to be first because 'field_origin' can't be None
    # when using 'issubclass'.
    if field_origin is None:
        return _handle_basic_types(field, field_type, value, type_map)

    # Handle mappings (e.g., dicts)
    if issubclass(field_origin, Mapping):
        return _handle_mapping_types(field, field_type, field_origin, value, type_map)

    # Handle tuples separately because the typing is idiosyncratic.
    if issubclass(field_origin, tuple):
        return _handle_tuple_types(field, field_type, value, type_map)

    # Handle other iterables
    if issubclass(field_origin, Iterable) and not isinstance(value, str):
        return _handle_iterable_types(field, field_type, field_origin, value, type_map)

    raise ValueError(
        f"Field '{field}' has an unsupported container type: {field_origin}"
    )


def _handle_union_type(
    field: Field, field_type: Any, value: Any, type_map: dict[str, Any]
) -> Any:
    field_args = get_args(field_type)

    if NoneType in field_args and value is None:
        return None

    for arg in field_args:
        try:
            return _handle_field(field, arg, value, type_map)
        except (ValueError, TypeError):
            continue

    raise ValueError(
        f"Field '{field.name}' with value '{value}' cannot be converted to any type in "
        f"{field_args}."
    )


def _handle_basic_types(
    field: Field, field_type: Any, value: Any, type_map: dict[str, Any]
) -> Any:
    # If the value is already an instance of the field type, we can just
    # return it.
    if isinstance(value, field_type):
        return value

    if isinstance(value, str) and issubclass(
        field_type, (datetime.date, datetime.datetime)
    ):
        return parse_date(value)  # noqa

    if is_dataclass(field_type) and isinstance(value, dict):
        return construct_dataclass(value, cast(Type, field_type), type_map)

    raise TypeError(
        f"Field '{field.name}' expected type '{field_type}', but got '{type(value)}"
    )


def _handle_mapping_types(
    field: Field, field_type: Any, field_origin, value: Any, type_map: dict[str, Any]
) -> Any:
    if not isinstance(value, Mapping):
        raise TypeError(
            f"Field '{field.name}' expected type '{field_type}', but got '{type(value)}"
        )

    field_args = get_args(field_type)
    key_type = field_args[0] if field_args else Any
    value_type = field_args[1] if field_args and len(field_args) > 1 else Any

    return field_origin(
        (
            _handle_field(field, key_type, k, type_map),
            _handle_field(field, value_type, v, type_map),
        )
        for k, v in value.items()
    )


def _handle_tuple_types(
    field: Field, field_type: Any, value: Any, type_map: dict[str, Any]
) -> tuple:
    if not isinstance(value, Iterable):
        raise TypeError(
            f"Field '{field.name}' expected type '{field_type}', but got {type(value)}"
        )

    values = list(value)
    n_values = len(values)
    field_args = get_args(field_type)

    ellipsis_index = field_args.index(Ellipsis) if Ellipsis in field_args else -1
    if not field_args or ellipsis_index == 0:
        # Note tuple[...] does not seem to be valid (pydantic will complain)
        # Can be any length and type
        inner_types = [Any for _ in range(n_values)]
    elif ellipsis_index > 0:
        # Can be any length
        inner_types = [
            field_args[i] if i < ellipsis_index else field_args[ellipsis_index - 1]
            for i in range(n_values)
        ]
    else:
        # Must be a specific length and specific types
        if n_values != len(field_args):
            raise ValueError(
                f"Field '{field.name}' expected length {n_values} but got {len(field_args)}"
            )
        inner_types = field_args

    return tuple(
        _handle_field(field, t, e, type_map) for t, e in zip(inner_types, values)
    )


def _handle_iterable_types(
    field: Field,
    field_type: Any,
    field_origin: Any,
    value: Any,
    type_map: dict[str, Any],
) -> Iterable:
    if not isinstance(value, Iterable):
        raise TypeError(
            f"Field '{field.name}' expected type '{field_type}', but got {type(value)}"
        )

    field_args = get_args(field_type)
    inner_type = field_args[0] if field_args else Any

    return field_origin(
        _handle_field(field, inner_type, item, type_map) for item in value
    )


# endregion Construct Dataclass


class DataclassRegistry[D]:
    """A class for registering and managing a collection of dataclasses."""

    def __init__(self: Self):
        """Initialize the registry."""
        self.registered_classes: set[Type[D]] = set()

    def __iter__(self: Self) -> Iterator[Type[D]]:
        return iter(self.registered_classes)

    def __len__(self) -> int:
        return len(self.registered_classes)

    def register(self, *classes: Type[D]):
        """
        Add a class to the registry.

        Parameters
        ----------
        classes : Type[D]
            The class or classes to add.
        """
        for cls in classes:
            self.registered_classes.add(cls)

    def construct_registered_dataclass(
        self, data: Mapping[str, Any], type_map: Optional[dict[str, Any]] = None
    ) -> D:
        """Try to construct one of the registered dataclasses from the given data.

        Parameters
        ----------
        data : Mapping[str, Any]
            The data to use for constructing a dataclass.
        type_map : dict[str, Any], optional
            Additional types that should be made available (imported) when
            resolving type hints for the registered dataclasses.

        Returns
        -------
        D
            An instance of a registered dataclass that is compatible with the
            given data.

        Raises
        ------
        icontract.ViolationError
            If ``data`` does not contain values for all required attributes
            of the dataclass (or any nested dataclass).
        icontract.ViolationError
            If ``data`` is not a :class:`~typing.Mapping`.
        ValueError
            If a registered dataclass cannot be constructed from the data.

        """
        successful_constructions = []

        for dataclass_type in self.registered_classes:
            try:
                instance = construct_dataclass(data, dataclass_type, type_map)
            except (ValueError, TypeError, ViolationError):
                continue
            else:
                successful_constructions.append(instance)

        if len(successful_constructions) == 0:
            raise ValueError(
                "The 'data' cannot be used to construct any registered dataclasses."
            )

        if len(successful_constructions) > 1:
            dataclass_names = [
                type(instance).__name__ for instance in successful_constructions
            ]

            raise ValueError(
                f"The 'data' is ambiguous and can be used to construct multiple dataclasses: "
                f"{dataclass_names}."
            )

        return successful_constructions[0]


class FrozenDataclassMeta(type):
    """Turn the class into a ``pydantic`` dataclass if it is not one already.

    .. note::
        Although this will turn any class that uses it into a ``pydantic``
        dataclass, it will cause most linters (including PyCharm)
        to issue unresolved attribute warnings for all attributes.
    """

    def __new__(
        cls: Type[FrozenDataclassMeta],
        name: str,
        bases: tuple[Type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> Type:
        new_cls = super().__new__(cls, name, bases, namespace, **kwargs)
        new_cls = dataclass(new_cls, frozen=True)  # noqa

        return new_cls


class EnforceFrozenDataclassMeta(type):
    """Requires that the class must be a standard or pydantic dataclass.

    Any class using this metaclass is not a frozen dataclass will raise a
    ``TypeError`` exception.

    .. note::
        This is not particularly useful because the metaclass will be
        applied before any decorators, so even if a class is decorated
        with @dataclass it will still raise a TypeError here, because
        it has not been applied yet.
    """

    def __new__(
        cls: Type[EnforceFrozenDataclassMeta],
        name: str,
        bases: tuple[Type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> EnforceFrozenDataclassMeta:
        new_cls = super().__new__(cls, name, bases, namespace, **kwargs)

        if not is_frozen(new_cls):
            raise TypeError(
                f"{name} must be decorated with '@dataclasses.dataclass(frozen=True)' or "
                f"'@pydantic.dataclasses.dataclass(frozen=True)'"
            )

        return new_cls



"""An implementation of Bayesian t-tests."""

# Future Library
from __future__ import annotations

# Standard Library
import logging

from dataclasses import dataclass
from typing import Any, Optional, Union

# 3rd Party Library
import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pymc as pm

from icontract import require
from matplotlib.axes import Axes
from numpy.typing import ArrayLike

# 1st Party Library
from rwskit.benchmarking import is_iterable

log = logging.getLogger(__name__)
logging.getLogger("PIL").setLevel("INFO")
logging.getLogger("matplotlib").setLevel("INFO")


InferenceData = Union[az.InferenceData, pm.backends.base.MultiTrace]  # type: ignore
"""
Data generated during Bayesian inference.
"""


# region Validation Functions
def _is_number(value: Any) -> bool:
    return isinstance(value, (int, float, np.floating, np.integer))


def _is_numeric_iterable(obj: Any) -> bool:
    return is_iterable(obj) and all(_is_number(x) for x in obj)


# endregion Validation Functions


@dataclass(frozen=True)
class BayesFactor:
    """
    The pair of Bayes Factors for two models :math:`M_0` and :math:`M_1`.
    Both directions are provided for convenience even though one can always be
    derived from the other.

    """

    bf_10: float
    """
    The Bayes Factor :math:`BF_{10} = \\frac{p(y | M_1)}{p(y | M_0)}`.
    """

    inference_data: Optional[InferenceData] = None
    """
    The trace of the sampling process, i.e., all the samples generated and
    used to compute the Bayes Factors.
    """

    @property
    def bf_01(self) -> float:
        """
        The Bayes Factor :math:`BF_{01} = \\frac{1}{BF_{10}} = \\frac{p(y | M_0)}{p(y | M_1)}`.
        """
        return 1.0 / self.bf_10

    def is_significant(self, significance_level: float = 10.0) -> bool:
        """Test if support for the hypothesis is significant.

        Returns ``True`` if the Bayes Factor in support of the hypothesis
        model (i.e., :math:`BF_{10}`) is greater than the specified
        ``significance_level``. See `this article <https://www.statology.org/bayes-factor/>`__
        for a table of common significance levels.

        If :data:`bf_10` is ``inf``, this function will always return ``True``.

        Parameters
        ----------
        significance_level : float, default = 10.0
            The Bayes Factor to consider evidence in favor of the hypothesis
            significant.

        Returns
        -------
        bool
            ``True`` if :data:`bf_10` is greater than the specified ``significance_level``.
        """
        if np.isinf(self.bf_10):
            return True

        return self.bf_10 > significance_level

    def plot_trace(self, combined: bool = True, show: bool = False) -> Optional[Axes]:
        """Plot the inference trace using the :data:`inference_data`.

        This method fixes the title formatting of :func:`arviz.plot_trace`.
        Without this fix, the subplot titles on the left graph overlap
        with the x-axis of the subplot just above it making them
        illegible.

        Parameters
        ----------
        combined : bool
            If ``True``, the inference chains will all be plotted using a
            single series.
        show : bool
            If ``True`` render the plot to the current device.

        Returns
        -------
        Axes
            The matplotlib axes.
        """
        if self.inference_data is None:
            log.warning("There is no inference data to use for plotting.")
            return None

        ax = az.plot_trace(self.inference_data, combined=combined)
        fig = ax[0, 0].get_figure()
        fig.subplots_adjust(hspace=0.5)

        if show:
            plt.show()

        return ax


@require(
    lambda group_1: _is_numeric_iterable(group_1),
    "'group_1' must be a sequence of numbers",
)
@require(
    lambda group_2: _is_numeric_iterable(group_2),
    "'group_2' must be a sequence of numbers",
)
@require(
    lambda beta_0_range: (
        True
        if beta_0_range is None
        else len(beta_0_range) == 2 and beta_0_range[0] < beta_0_range[1]
    ),
    "'beta_0_range[0]' must be less than 'beta_0_range[1]'",
)
@require(
    lambda cohen_d_scale: cohen_d_scale > 0, "'cohen_d_scale' must be a positive number"
)
@require(lambda sigma_scale: sigma_scale > 0, "'sigma_scale' must be a positive number")
@require(
    lambda n_posterior_samples: isinstance(n_posterior_samples, int)
    and n_posterior_samples > 0,
    "'n_posterior_samples' must be a positive integer",
)
@require(
    lambda n_prior_samples: isinstance(n_prior_samples, int) and n_prior_samples > 0,
    "'n_prior_samples' must be a positive integer",
)
@require(
    lambda random_seed: (
        isinstance(random_seed, int) if random_seed is not None else None
    ),
    "'random_seed' must be an integer",
)
def savage_dickey_t_test(
    group_1: ArrayLike,
    group_2: ArrayLike,
    beta_0_range: Optional[tuple[float, float]] = None,
    cohen_d_scale: float = 1,
    sigma_scale: float = 10,
    n_posterior_samples: int = 1000,
    n_prior_samples: int = 4000,
    random_seed: Optional[int] = None,
    return_trace: bool = False,
) -> BayesFactor:
    """Perform a Bayesian t-test on the two arrays.

    The test can be used to estimate the likelihood that the true means of two
    groups are different. The returned Bayes Factor represents the ratio between
    the probability of the null-hypothesis :math:`H_{0}`, that the means are
    equal, and the alternate hypothesis :math:`H_{1}`, the means are different.

    :math:`BF_{10} > 1` indicates support that the means are different.
    :math:`BF_{10} > 10` generally indicates strong support that the means
    are different (:math:`H_{1}` is more than 10 times as likely than
    :math:`H_{0}`). Note, :math:`BF_{10} = \\frac{1}{BF_{01}}`.

    The default priors are extremely weakly informative. In practice, this
    means more data is needed to make strong conclusions about the data. If
    more information is known about the groups it should be used to
    refine the priors.

    See the description of :ref:`model` below for the meaning of the parameters:
    ``beta_0``, ``cohen_d``, and ``sigma``.

    Parameters
    ----------
    group_1 : array-like
        The first array.
    group_2 : array_like
        The second array
    beta_0_range : tuple[float, float], optional
        The range of values for the Uniform prior on :math:`\\beta_{1}`. If
        not provided, defaults to
        ``(min(group_1 + group_2), max(group_1 + group_2))``.
    cohen_d_scale : float
        The scale to use for the ``cohen_d`` prior drawn from a Cauchy
        distribution.
    sigma_scale : float
        The scale to use for model noise drawn from a half Normal
        distribution. Using a large value will bias the model towards
        indicating the groups are the same (because the prior belief
        is most of the variability is due to random noise).
    n_posterior_samples : int, default = 1000
        The number of samples to draw from the posterior distribution.
    n_prior_samples : int, default = 4000
        The number of samples to draw from the predictive prior
        distribution.
    random_seed : int, default = None
        The random seed to use when drawing samples.
    return_trace: bool, default = False
        If ``True`` return the sampled inference data.

    Returns
    -------
    BayesFactor
        The Bayes Factors for :math:`H_{10}` and :math:`H_{01}`. When the
        groups are very different sometimes the posterior of the fit model
        does not contain the reference value for comparison (i.e., the value
        ``0``). In this case ``bf_10=np.inf`` and ``bf_01=0``. If this behavior
        is not caused by an error in the sampling process, it provides extreme
        support that the group means are different (but probably not infinite
        support).


    .. _model:

    The Model
    ---------
    The test is implemented based on the method proposed by
    `Rouder et al. (2009) <https://pubmed.ncbi.nlm.nih.gov/19293088/>`__
    and described by
    `Maarten Speekenbrink <https://mspeekenbrink.github.io/sdam-book/ch-Bayes-factors.html#a-bayesian-t-test>`__.

    The model is defined as:

    .. math::

       Y            &\\sim \\mathcal{N}(\\mu, \\sigma_{y}) \\\\
       \\sigma_{y}  &\\sim \\mathcal{N^{+}}(\\tau_{\\sigma_{y}}) \\\\
       d_{c}        &\\sim \\text{Cauchy}(0, \\tau_{d_{c}}) \\\\
       \\beta_{1}   &= d_{c} \\cdot \\sigma_{y} \\\\
       \\beta_{0}   &\\sim \\text{Uniform}(\\tau_{\\beta_{0}[0]}, \\tau_{\\beta_{0}[1]}) \\\\
       \\mu         &= \\beta_{0} \\cdot \\beta_{1} \\cdot C_{h}(X) \\\\

    :math:`\\sigma_{y}` is the random noise in the process, i.e., the
    expected standard deviation of the values from the mean du to chance.
    The prior is drawn from a half Normal distribution with scale
    :math:`\\tau_{\\sigma_{y}}`.

    :math:`X` is an indicator variable for our two groups. It is
    coded using Helmert coding, :math:`C_{h}(X)`, which compares a group
    mean to the average of the other group means. In the case of 2 groups, it
    will cause :math:`\\beta_{1}` to converge to :math:`\\mu_{1} - \\mu_{2}`
    and :math:`\\beta_{0}` to :math:`\\frac{\\mu_{1} + \\mu_{2}}{2}`.

    Because the range of values of
    :math:`\\mu_{1} - \\mu_{2}` is highly dependent data generating process, it
    is difficult to pick a prior that is likely to work well for any 2 groups
    we might encounter. The proposed method is to sample from
    `Cohen's d <https://statisticsbyjim.com/basics/cohens-d/>`__,
    :math:`d_{c} = \\frac{\\mu_1 - \\mu_{2}}{\\sigma_{y}}`, instead.
    :math:`\\beta_{1}` can be easily recovered by multiplying
    :math:`d_{c}` by :math:`\\sigma_{y}`.

    :math:`\\beta_{0}` is drawn from a Normal distribution with scale
    :math:`\\tau_{\\beta_{0}}`.

    :math:`\\mu` is the linear model using the indicator variable defined above.
    The intercept is :math:`\\beta_{0}` and the slope is :math:`\\beta_{1}`.

    :math:`Y \\sim \\mathcal{N}(\\mu, \\sigma_{y})` is the full model that
    incorporates the random noise.

    Hypothesis Testing
    ------------------

    We can use the model for hypothesis testing by setting up our two
    hypotheses as follows:

    .. math::

       H_0: \\mu_1 - \\mu_2 &=  \\beta_{1} &=       0   \\\\
       H_1: \\mu_1 - \\mu_2 &=  \\beta_{1} &\\neq   0

    Because the models for :math:`H_{0}` and :math:`H_{1}` are
    properly nested [1]_, we can use the
    `Savage-Dickey density ratio <https://statproofbook.github.io/P/bf-sddr.html>`__
    to estimate the Bayes Factor. So that:

    .. math::

       \\text{BF}_{01} = \\frac{p(\\beta_{1}=0 | D, Y)}{p(\\beta_{1}=0 | Y)}

    Where :math:`D` is our observed data and :math:`Y` is our (linear) model.
    In words this essentially says, the Bayes Factor in support of the
    null-hypothesis can be found by taking the ratio of the probability that
    :math:`\\beta_{1}` equals 0 in the posterior distribution to the
    probability that :math:`\\beta_{1}` equals 0 in the (predictive) prior
    distribution.

    .. [1] Briefly, two models are properly nested if they have the same
           parameters, but in one model, :math:`M_{1}`, they are all free
           to take on any value while in the other (at least) one of the
           parameters is fixed to a constant value.
    """
    # Create the observed values by concatenating the lists
    group_1 = np.asarray(group_1)
    group_2 = np.asarray(group_2)

    observed = np.concatenate((group_1, group_2))

    # Set up the Helmert contrast coded predictor variables.
    x = [0.5] * len(group_1) + [-0.5] * len(group_2)

    if not beta_0_range:
        # This is not entirely kosher, but seems reasonable for the task.
        # beta_1 cannot be below or above the min and max of the observed data.
        beta_0_range = (np.min(observed), np.max(observed))

    beta_0_lower, beta_0_upper = beta_0_range

    with pm.Model():
        # Depending on the data, calculating the Bayesian Factor (below) will
        # issue a warning that the reference value is outside the posterior.
        # I think this is normal, and as the warning indicates, this just means
        # the model is very sure the two groups do not have the same mean.

        # 'sigma' represents the random noise in the data.
        sigma = pm.HalfNormal("sigma", sigma=sigma_scale)

        # 'beta_0' represents the y-intercept and should converge to the mean
        # of means: (mu_1 + mu_2) / 2. 'beta_0' is drawn from a (very) weakly
        # informative prior.
        beta_0 = pm.Uniform("beta_0", lower=beta_0_lower, upper=beta_0_upper)

        # 'cohen_d' represents Cohen's d. Cohen's d is defined as
        # (mu_1 - mu_2) / sigma_y. We define the prior on Cohen's d,
        # instead of directly on (mu_1 - mu_2), because it is a normalized
        # value, which makes it invariant to the scale of the input data.
        cohen_d = pm.Cauchy("cohen_d", alpha=0, beta=cohen_d_scale)

        # 'beta_1', under the specified contrast-coded scheme, represents
        # the difference in group means (mu_1 - mu_2), and is the value we
        # are interested in comparing.
        beta_1 = pm.Deterministic("beta_1", cohen_d * sigma)  # noqa

        # The linear equation
        mu = beta_0 + beta_1 * x  # noqa

        # The final distribution
        pm.Normal("y", mu=mu, sigma=sigma, observed=observed)

        # Draw samples from the posterior
        idata = pm.sample(
            draws=n_posterior_samples, random_seed=random_seed, progressbar=False
        )

        # Draw samples from the prior (predictive) distribution.
        # The Bayes Factor (under the nested model assumptions) will be the
        # ratio of the values posterior to prior when beta_1 == 0
        idata.extend(
            pm.sample_prior_predictive(draws=n_prior_samples, random_seed=random_seed)
        )

    reference_var = "beta_1"
    reference_val = 0

    # Check to see if the reference value is outside the posterior.
    posterior = az.data.utils.extract(idata, var_names=reference_var)
    if not posterior.min() <= reference_val <= posterior.max():
        # The reference value was never sampled from the posterior
        # distribution. This usually indicates that the reference value
        # is extremely unlikely given the data. The Bayes Factor is
        # probably not actually infinity, but we don't really have a better
        # estimate. In our case it shouldn't really matter, because it
        # just means the null-hypothesis is (almost) certainly false.
        return BayesFactor(np.inf)

    bf, ax = az.plot_bf(
        idata, var_name=reference_var, ref_val=reference_val, show=False
    )
    ax.remove()

    inference_data = idata if return_trace else None
    return BayesFactor(bf["BF10"], inference_data=inference_data)  # type: ignore



"""
Numpy utilities.
"""

# Future Library
from __future__ import annotations

# Standard Library
import datetime
import logging

from typing import Any, Callable, Literal, Optional

# 3rd Party Library
import numpy as np
import pandas as pd

# 1st Party Library
from rwskit.collections_ import get_first_non_null_value, is_iterable

log = logging.getLogger(__name__)


CompleteCheckType = Literal["finite", "nan"]
"""
The valid values for the ``check_type`` parameter of :meth:`is_complete`.
"""

_python_to_dtype_map = {
    int: np.int_,
    bool: np.bool_,
    float: np.float64,
    complex: np.complex128,
    bytes: np.bytes_,
    str: np.str_,
    datetime.datetime: np.datetime64,
    datetime.date: np.datetime64,
    pd.Timestamp: np.datetime64,
}


def get_dtype(obj: Optional[Any]) -> Optional[np.dtype]:
    """
    Return a dtype that can be used to represent an arbitrary object.

    For objects that are already numpy arrays or scalars, the dtype from that
    object is returned. For other python objects, a set of heuristics are used.
    If the object is iterable, the method will try to determine the type of
    object

    If the dtype can't be determined, either because the input object was
    ``None``, or it is a collection that does not contain any non-null
    values, then the method will return ``None``.

    Parameters
    ----------
    obj : Any
        An object.

    Returns
    -------
    numpy.dtype, optional
        A dtype that can be used to represent an arbitrary object, or ``None``
        if the input object is ``None``.

    """
    if obj is None:
        return None

    if isinstance(obj, str) or not is_iterable(obj):
        # Try to use numpy to automatically detect the dtype
        dtype = np.array(obj).dtype

        # We've found a specific dtype so return it.
        if dtype != np.object_:
            return dtype

        return _python_to_dtype_map.get(type(obj), np.object_)
    else:
        # Get the first non-null value in the iterable
        a_value = get_first_non_null_value(obj)

        # Recurse on the fetched object.
        return get_dtype(a_value)


def is_structured(array: np.ndarray) -> bool:
    """
    Checks if a numpy array is a structured array or not.

    Parameters
    ----------
    array : np.ndarray
        The array to check.

    Returns
    -------
    bool
        ``True`` if the array is a structured array, ``False`` otherwise.
    """
    return array.dtype.names is not None


# region is_complete
def is_complete(array: np.ndarray, check_type: CompleteCheckType = "finite") -> bool:
    """
    An array is complete if all its numeric values are finite and all other
    values are not ``None``.

    .. note::
        For arrays with ``dtype=np.object_``, the method only checks to make sure
        no value in the array is ``None``.

    Parameters
    ----------
    check_type : {"finite", "nan"}
        Determines how completeness is defined.

        nan
            Requires that no values are ``np.NaN`` or ``None``.
        finite
            In addition to the criteria for ``nan`` all values must also be
            finite (i.e., not ``np.inf`` or ``-np.inf``).

    array : np.ndarray
        The array to check.

    Returns
    -------
    bool
        ``True`` if the array is complete, ``False`` otherwise.
    """
    complete_fn = _get_complete_function(check_type, array)

    if is_structured(array):
        return _is_structured_complete(array, complete_fn)

    return _is_unstructured_complete(array, complete_fn)


def _get_complete_function(
    check_type: CompleteCheckType, array: np.ndarray
) -> Callable[[np.ndarray], bool]:
    # np.dot is always fastest for 1D arrays.
    if check_type == "finite" and array.ndim <= 1:
        return lambda a: np.isfinite(np.dot(a, a))

    # When checking for finite values, we have to use an accumulating
    # function, otherwise the possibly infinite values don't propagate.
    if check_type == "finite":
        return lambda a: np.isfinite(np.sum(a))

    if array.ndim <= 1:
        return lambda a: not np.isnan(np.dot(a, a))

    # When checking for NaNs in greater than 1D creating a boolean
    # mask is fastest for small N, but eventually np.sum is faster.
    if array.size < 1500000:
        return lambda a: not np.any(np.isnan(a))

    return lambda a: np.isnan(np.sum(a))


def _is_unstructured_complete(
    array: np.ndarray,
    complete_fn: Callable[[np.ndarray], bool],
) -> bool:
    # Check if floating point values are finite.
    if np.issubdtype(array.dtype, np.floating):
        return complete_fn(array)

    # Check if object types are non-null
    if np.issubdtype(array.dtype, np.object_):
        return None not in array

    # All other values must have a value
    return True


def _is_structured_complete(
    array: np.ndarray,
    complete_fn: Callable[[np.ndarray], bool],
) -> bool:
    return all(
        _is_unstructured_complete(array[c], complete_fn) for c in array.dtype.names
    )


# endregion is_complete


# region finite_cases
def finite_cases(array: np.ndarray) -> np.ndarray:
    """
    Return only the rows whose values are finite and not ``None``.

    Parameters
    ----------
    array : np.ndarray
        The input array with possibly invalid values.

    Returns
    -------
    np.ndarray
        An array with only non-null finite values.
    """
    if is_structured(array):
        return _structured_finite_cases(array)

    return _unstructured_finite_cases(array)


def _unstructured_finite_cases(array: np.ndarray) -> np.ndarray:
    return array[np.isfinite(array).all(axis=1)]


def _structured_finite_cases(array: np.ndarray) -> np.ndarray:
    masks = []

    for field in array.dtype.names:
        column = array[field]

        if np.issubdtype(column.dtype, np.floating):
            masks.append(np.isfinite(column))
        elif np.issubdtype(column.dtype, np.object_):
            try:
                masks.append(np.isfinite(column.astype(float)))
            except ValueError:
                masks.append(column != None)  # noqa

    mask = np.logical_and(*masks)

    return array[mask]


# endregion finite_cases


# region group_by
def group_by(array: np.ndarray, column: str | int) -> list[np.ndarray]:
    """
    Group an ``array`` by the values in the given ``column``. If ``array`` is
    as structured array, then ``column`` should be the name of one of the
    fields, otherwise it should be the index of the column you want to use.

    Parameters
    ----------
    array : np.ndarray
        The array to group.
    column : str | int
        If ``array`` is structured then ``column`` should be the name of the
        column to group by. Otherwise, it should be the index of the column.

    Returns
    -------
    list
        Returns the group as a list of arrays.
    """
    # Based on vincent's method: https://stackoverflow.com/a/43094244
    if is_structured(array):
        # The indices such that the rows would be sorted by the values in
        # the given `column`.
        indices = array[column].argsort()

        # The input array sorted by those indices.
        sorted_array = array[indices]

        # When `return_index=True` np.unique will return two arrays. The
        # first will be the actual unique values. The second will be
        # the indices in the original array that point to the first instance
        # of each unique value. To split the array into groups we want all
        # but the first value of the second array.
        group_indices = np.unique(sorted_array[column], return_index=True)[1][1:]
    else:
        if not isinstance(column, int):
            raise ValueError("'column' must be an integer for unstructured arrays.")

        # The indices such that the rows would be sorted by the values in
        # the given `column`.
        indices = array[:, column].argsort()

        # The input array sorted by those indices.
        sorted_array = array[indices]

        # When `return_index=True` np.unique will return two arrays. The
        # first will be the actual unique values. The second will be
        # the indices in the original array that point to the first instance
        # of each unique value. To split the array into groups we want all
        # but the first value of the second array.
        group_indices = np.unique(sorted_array[:, column], return_index=True)[1][1:]

    # Get the groups
    return np.split(sorted_array, group_indices)


# endregion group_by



# Future Library
from __future__ import annotations

# Standard Library
import os

from importlib.resources.abc import Traversable
from typing import Union

PathLike = Union[str, os.PathLike, Traversable]



"""Utilities related to exceptions."""

# Standard Library
from typing import Any, Type


def raise_exception(
    msg: str, exception_type: Type[Exception] = ValueError, **kwargs: Any
):
    """Always raises the given ``exception_type` with the supplied ``msg`` and ``kwargs``.

    Parameters
    ----------
    msg : str
        The message for the exception.
    exception_type : Type[Exception], default = ``ValueError``
        The type of exception to raise.
    kwargs : Any
        Any additional kwargs to pass to the exception

    Raises
    ------
    exception_type
    """
    raise exception_type(msg, **kwargs)



"""Hash utilities."""

# Future Library
from __future__ import annotations

# Standard Library
import logging

from typing import Any, Callable, Literal, Optional

# 3rd Party Library
import xxhash

from icontract import require
from msgspec import msgpack

# 1st Party Library
from rwskit.collections_ import recursive_sort
from rwskit.numeric import to_signed

log = logging.getLogger(__name__)

HashSize = Literal[32, 64, 128]


class ObjectHasher:
    """Hash objects using xxHash.

    The only requirement is that the object be ``msgpack`` serializable.
    """

    _hashers: dict[HashSize, Callable] = {
        32: xxhash.xxh32,
        64: xxhash.xxh3_64,
        128: xxhash.xxh3_128,
    }

    @require(lambda hash_size: hash_size in (32, 64, 128))
    def __init__(self, hash_size: HashSize = 128, signed: bool = False):
        self._hash_size: HashSize = hash_size
        self._encoder = msgpack.Encoder()
        self._hasher = self._hashers[hash_size]
        self._signed = signed

    @property
    def hash_size(self) -> HashSize:
        """Get integer size of the returned hash values this hasher produces."""
        return self._hash_size

    @require(lambda self, signed: not signed or self._hash_size < 128)
    def hash(self, obj: Any, signed: Optional[bool] = None) -> int:
        """
        Hash the object using the hash size specified in the constructor.

        ``xxHash`` returns an unsigned value, but it can be converted to a
        signed value if the hash size is less than 128 bits.

        ..note::
            Only values supported by
            `msgpack <https://jcristharif.com/msgspec/supported-types.html>`__
            can be hashed.

        ..note::
            This can only hash integers up to 64-bits.

        Parameters
        ----------
        obj : Any
            The object to hash
        signed : bool, default = False
            Whether to convert the value to a signed integer using
            :meth:`~rwskit.numeric.to_signed`.

        Returns
        -------
        int
            An integer representing the hash of the object.

        Raises
        ------
        OverflowError
            If the data contains an integer that is outside the range
            -2^63 to 2^64-1.
        """
        signed = signed or self._signed
        data = self._encoder.encode(obj)
        unsigned_value = self._hasher(data).intdigest()

        return to_signed(unsigned_value) if signed else unsigned_value

    @require(lambda self, signed: not signed or self._hash_size < 128)
    def hash_sorted(self, obj: Any, signed: Optional[bool] = None) -> int:
        """Sort a collection using :func:`~rwskit.collections_.recursive_sort` and hash the result.

        This should provide a more robust hash that should return the same
        value for collections containing the same data, but in a different
        order.
        """

        signed = signed or self._signed
        return self.hash(recursive_sort(obj), signed=signed)
